import os
import sys
import logging
import platform
import socket
from argparse import ArgumentParser
import qrcode
from socketio import Server, WSGIApp
from eventlet import wsgi, listen
if platform.system() == 'Linux':
	try:
		from .nix.device import X360Device, DS4Device
		from .nix.setup import setup
	except ImportError:
		from j2dx.nix.device import X360Device, DS4Device
		from j2dx.nix.setup import setup
else:
	try:
		from .win.device import X360Device, DS4Device
		from .win.setup import setup
	except ImportError:
		from j2dx.win.device import X360Device, DS4Device
		from j2dx.win.setup import setup


def parse_args():
	parser = ArgumentParser()
	if platform.system() == 'Linux':
		parser.add_argument(
			'user',
			nargs='?',
			default=os.getenv('SUDO_USER'),
			help='Only used with --setup. User to configure for UInput access.'
		)
	parser.add_argument(
		'-s', '--setup',
		action='store_true',
		help='Setup system for virtual device creation.'
		'Must have root/administrator access.',
	)
	parser.add_argument(
		'-H', '--host',
		default=None,
		help='Hostname or IP address the server will listen on.',
	)
	parser.add_argument(
		'-p', '--port',
		type=int, default=8013,
		help='Port the server will listen on. Defaults to 8013.',
	)
	parser.add_argument(
		'-d', '--debug',
		action='store_true',
		help='Print debug information.',
	)
	return parser.parse_args()


def get_logger(debug):
	logging.basicConfig(
		level=logging.DEBUG if debug else logging.INFO)
	logging.getLogger('engineio.server').setLevel(
		logging.INFO if debug else logging.ERROR)
	logging.getLogger('socketio.server').setLevel(
		logging.INFO if debug else logging.ERROR)
	wsgi_logger = logging.getLogger('eventlet.wsgi.server')
	wsgi_logger.setLevel(
		logging.INFO if debug else logging.ERROR)
	return (logging.getLogger('J2DX.server'), wsgi_logger)


def default_host():
	sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
	try:
		sock.connect(('1.255.255.255', 1))
		IP = sock.getsockname()[0]
	except IndexError:
		IP = '127.0.0.1'
	finally:
		sock.close()
	return IP


def main():
	args   = parse_args()
	logger, wsgi_logger = get_logger(args.debug)

	if args.setup:
		if platform.system() == 'Linux':
			setup(args.user)
		else:
			setup()
		sys.exit(0)
		return

	CLIENTS = {}
	DEVICES = {}

	sio = Server(logger=args.debug, engineio_logger=args.debug)
	app = WSGIApp(sio)

	@sio.event
	def connect(sid, environ):
		CLIENTS[sid] = environ['REMOTE_ADDR']
		logger.info(f'Client connected from {environ["REMOTE_ADDR"]}')
		logger.debug(f'Client {environ["REMOTE_ADDR"]} sessionId: {sid}')

	@sio.event
	def intro(sid, data):
		if data['id'] == 'x360':
			DEVICES[sid] = X360Device(data['device'], CLIENTS[sid])
		else:
			DEVICES[sid] = DS4Device(data['device'], CLIENTS[sid])
		logger.info(
			f'Created virtual {data["type"]} gamepad for {data["device"]} '
			f'at {CLIENTS[sid]}')

	@sio.event
	def input(sid, data):
		logger.debug(f'Received INPUT event::{data["key"]}: {data["value"]}')
		DEVICES[sid].send(data['key'], data['value'])

	@sio.event
	def disconnect(sid):
		device = DEVICES.pop(sid)
		logger.info(f'Disconnected device {device.device} at {device.address}')
		device.close()

	try:
		host = args.host or default_host()
		sock = listen((host, args.port))
	except PermissionError:
		sys.exit(
			f'Port {args.port} is not available. '
			f'Please specify a different port with -p option.'
		)
	logger.info(f'Listening on http://{host}:{args.port}/')
	qr = qrcode.QRCode()
	qr.add_data(f'j2dx://{host}:{args.port}/')
	if platform.system() == 'Windows':
		import colorama
		colorama.init()
	qr.print_ascii(tty=True)
	wsgi.server(sock, app, log=wsgi_logger, log_output=args.debug, debug=False)


if __name__ == '__main__':
	main()



import os
import sys
import logging
from subprocess import run


logger = logging.getLogger('J2DX.Linux.Setup')


GROUP = 'j2dx'
UDEV_PATH = '/etc/udev/rules.d/80-j2dx.uinput.rules'
UDEV_RULE = (
	f'SUBSYSTEM=="misc", KERNEL=="uinput", MODE="0660", GROUP="{GROUP}"')


def groupadd():
	proc = run(['groupadd', '-f', GROUP], capture_output=True, text=True)
	if proc.returncode != 0:
		logger.critical(proc.stderr)
		sys.exit(proc.returncode)
	logger.info(proc.stdout)


def usermod(user):
	proc = run(
		['usermod', '-a', '-G', GROUP, user], capture_output=True, text=True)
	if proc.returncode != 0:
		logger.critical(proc.stderr)
		sys.exit(proc.returncode)
	logger.info(proc.stdout)


def setup(user):
	if os.geteuid() != 0:
		logger.critical('Setup needs to run as root.')
		sys.exit('Try running this command with sudo.')

	try:
		with open(os.open(UDEV_PATH, os.O_CREAT | os.O_WRONLY, 0o660), 'w') as fd:
			fd.write(UDEV_RULE)
	except OSError as e:
		logger.critical(e.strerror)
		sys.exit(e.errno)

	groupadd()
	usermod(user)
	logger.info(
		'Setup complete. You may need to reboot for UDEV rules to take effect.')



import sys
import logging
from urllib.request import urlopen
from tempfile import NamedTemporaryFile
from ctypes import windll, c_void_p, c_wchar_p, c_int


VIGEM_REL = 'https://github.com/ViGEm/ViGEmBus/releases'
VIGEM_URI = f'{VIGEM_REL}/download/v1.16.112/ViGEmBus_Setup_1.16.115.exe'


SE_ERR_ACCESSDENIED = 5
SUCCESS = lambda errno: errno > 32    # noqa


ShellExecuteW = windll.shell32.ShellExecuteW
ShellExecuteW.argtypes = (
	c_void_p, c_wchar_p, c_wchar_p, c_wchar_p, c_wchar_p, c_int)
ShellExecuteW.restype = c_int


logger = logging.getLogger('J2DX.Windows.Setup')


def setup():
	with NamedTemporaryFile(
			delete=False, prefix='ViGEmBus-', suffix='.exe') as tmp:
		logger.info('Downloading compatible ViGEmBus driver.')
		try:
			resp = urlopen(VIGEM_URI)
			tmp.write(resp.read())
			vigembus = tmp.name
		except OSError as e:
			logger.critical(f'Driver download failed. Error {e.errno}')
			logger.info(
				f'Please manually download latest driver from {VIGEM_REL}')
			sys.exit(e.errno)
	logger.info('Running setup. Please accept UAC prompt when it appears.')
	ret = ShellExecuteW(
		None, 'runas',
		vigembus, '/exenoui /quiet /promptrestart',
		None, 1)
	if SUCCESS(ret):
		logger.info(
			'Setup complete. '
			'You may need to reboot before you can use virtual device driver.')
		sys.exit(0)
	else:
		if ret == SE_ERR_ACCESSDENIED:
			logger.critical('Setup needs administrator access to install drivers.')
		else:
			logger.critical(f'Unexpected error: {ret}')
		sys.exit(ret)



import logging
from abc import ABC, abstractmethod
from functools import reduce
from .ViGEm import client as vigem


logger = logging.getLogger('J2DX.Windows')


class Device(ABC):

	@abstractmethod
	def __init__(self, device, addr):
		self.device = device
		self.address = addr
		self._client = vigem.alloc()

	def _create_device(self):
		error = vigem.VIGEM_ERRORS(vigem.connect(self._client))
		if error != vigem.VIGEM_ERRORS.VIGEM_ERROR_NONE:
			logger.critical(
				f'Virtual {self.type} device {self.device} at {self.address}::\
					Connection to ViGEm driver failed::{error.name}')
			raise Exception(error.name)
		error = vigem.VIGEM_ERRORS(vigem.target_add(self._client, self._target))
		if error != vigem.VIGEM_ERRORS.VIGEM_ERROR_NONE:
			logger.critical(
				f'Virtual {self.type} device {self.device} at {self.address}::\
					Adding target failed::{error.name}')
			raise Exception(error.name)

	def close(self):
		error = vigem.VIGEM_ERRORS(vigem.target_remove(self._client, self._target))
		if error != vigem.VIGEM_ERRORS.VIGEM_ERROR_NONE:
			logger.critical(
				f'Virtual {self.type} device {self.device} at {self.address}::\
					Removing target failed::{error.name}')
			raise Exception(error.name)
		vigem.target_free(self._target)
		vigem.disconnect(self._client)
		vigem.free(self._client)
		logger.debug(
			f'Destroyed virtual {self.type} device for {self.device} \
				at {self.address}')

	@abstractmethod
	def send(self, key, value):
		pass


class X360Device(Device):

	buttons = {
		'main-button': vigem.XUSB_BUTTON.XUSB_GAMEPAD_GUIDE,
		'back-button': vigem.XUSB_BUTTON.XUSB_GAMEPAD_BACK,
		'start-button': vigem.XUSB_BUTTON.XUSB_GAMEPAD_START,
		'left-stick-press': vigem.XUSB_BUTTON.XUSB_GAMEPAD_LEFT_THUMB,
		'right-stick-press': vigem.XUSB_BUTTON.XUSB_GAMEPAD_RIGHT_THUMB,
		'left-bumper': vigem.XUSB_BUTTON.XUSB_GAMEPAD_LEFT_SHOULDER,
		'right-bumper': vigem.XUSB_BUTTON.XUSB_GAMEPAD_RIGHT_SHOULDER,
		'up-button': vigem.XUSB_BUTTON.XUSB_GAMEPAD_DPAD_UP,
		'right-button': vigem.XUSB_BUTTON.XUSB_GAMEPAD_DPAD_RIGHT,
		'down-button': vigem.XUSB_BUTTON.XUSB_GAMEPAD_DPAD_DOWN,
		'left-button': vigem.XUSB_BUTTON.XUSB_GAMEPAD_DPAD_LEFT,
		'y-button': vigem.XUSB_BUTTON.XUSB_GAMEPAD_Y,
		'x-button': vigem.XUSB_BUTTON.XUSB_GAMEPAD_X,
		'a-button': vigem.XUSB_BUTTON.XUSB_GAMEPAD_A,
		'b-button': vigem.XUSB_BUTTON.XUSB_GAMEPAD_B,
	}
	triggers = {
		'left-trigger': 'bLeftTrigger',
		'right-trigger': 'bRightTrigger',
	}
	axes_vertical = {
		'left-stick-Y': 'sThumbLY',
		'right-stick-Y': 'sThumbRY',
	}
	axes_horizontal = {
		'left-stick-X': 'sThumbLX',
		'right-stick-X': 'sThumbRX',
	}

	def __init__(self, device, addr):
		super().__init__(device, addr)
		self.type = "Xbox 360 Controller"
		self._target = vigem.target_x360_alloc()
		self._wButtons = set()
		self._report = vigem.XUSB_REPORT(
			wButtons=0,
			bLeftTrigger=0,
			bRightTrigger=0,
			sThumbLX=0,
			sThumbLY=0,
			sThumbRX=0,
			sThumbRY=0,
		)
		self._create_device()

	def send(self, key, value):
		if key in self.buttons:
			if value:
				self._wButtons.add(self.buttons[key])
			else:
				self._wButtons.remove(self.buttons[key])
		elif key in self.triggers:
			if value:
				setattr(self._report, self.triggers[key], vigem.XUSB_TRIGGER_MAX)
				logger.debug(f'{self.triggers[key]}::{vigem.XUSB_TRIGGER_MAX}')
			else:
				setattr(self._report, self.triggers[key], 0)
				logger.debug(f'{self.triggers[key]}::0')
		elif key in self.axes_vertical:
			axis = -round(value * vigem.XUSB_THUMB_MAX)
			setattr(self._report, self.axes_vertical[key], axis)
			logger.debug(f'{self.axes_vertical[key]}::{axis}')
		elif key in self.axes_horizontal:
			axis = round(value * vigem.XUSB_THUMB_MAX)
			setattr(self._report, self.axes_horizontal[key], axis)
			logger.debug(f'{self.axes_horizontal[key]}::{axis}')
		wButtons = reduce(lambda a, b: a | b, self._wButtons, 0)
		self._report.wButtons = wButtons
		logger.debug(
			f'wButtons::Mask {wButtons}::{[b.name for b in self._wButtons]}')
		error = vigem.VIGEM_ERRORS(
			vigem.target_x360_update(self._client, self._target, self._report)
		)
		if error != vigem.VIGEM_ERRORS.VIGEM_ERROR_NONE:
			logger.error(
				f'Virtual {self.type} device {self.device} at {self.address}::\
					Updating device state failed::{error.name}'
			)


class DS4Device(Device):

	buttons = {
		'back-button': vigem.DS4_BUTTONS.DS4_BUTTON_SHARE,
		'start-button': vigem.DS4_BUTTONS.DS4_BUTTON_OPTIONS,
		'left-stick-press': vigem.DS4_BUTTONS.DS4_BUTTON_THUMB_LEFT,
		'right-stick-press': vigem.DS4_BUTTONS.DS4_BUTTON_THUMB_RIGHT,
		'left-bumper': vigem.DS4_BUTTONS.DS4_BUTTON_SHOULDER_LEFT,
		'right-bumper': vigem.DS4_BUTTONS.DS4_BUTTON_SHOULDER_RIGHT,
		'y-button': vigem.DS4_BUTTONS.DS4_BUTTON_TRIANGLE,
		'x-button': vigem.DS4_BUTTONS.DS4_BUTTON_SQUARE,
		'a-button': vigem.DS4_BUTTONS.DS4_BUTTON_CROSS,
		'b-button': vigem.DS4_BUTTONS.DS4_BUTTON_CIRCLE,
	}
	specials = {
		'main-button': vigem.DS4_SPECIAL_BUTTONS.DS4_SPECIAL_BUTTON_PS,
	}
	dpad = {
		'up-button': vigem.DS4_DPAD_DIRECTIONS.DS4_BUTTON_DPAD_NORTH,
		'right-button': vigem.DS4_DPAD_DIRECTIONS.DS4_BUTTON_DPAD_EAST,
		'down-button': vigem.DS4_DPAD_DIRECTIONS.DS4_BUTTON_DPAD_SOUTH,
		'left-button': vigem.DS4_DPAD_DIRECTIONS.DS4_BUTTON_DPAD_WEST,
	}
	triggers = {
		'left-trigger': ('bTriggerL', vigem.DS4_BUTTONS.DS4_BUTTON_TRIGGER_LEFT),
		'right-trigger': ('bTriggerR', vigem.DS4_BUTTONS.DS4_BUTTON_TRIGGER_RIGHT),
	}
	axes_vertical = {
		'left-stick-Y': 'bThumbLY',
		'right-stick-Y': 'bThumbRY',
	}
	axes_horizontal = {
		'left-stick-X': 'bThumbLX',
		'right-stick-X': 'bThumbRX',
	}

	def __init__(self, device, addr):
		super().__init__(device, addr)
		self.type = "Sony Computer Entertainment Wireless Controller"
		self._target = vigem.target_ds4_alloc()
		self._wButtons = set()
		self._report = vigem.DS4_REPORT(
			bThumbLX=0,
			bThumbLY=0,
			bThumbRX=0,
			bThumbRY=0,
			wButtons=0,
			bSpecial=0,
			bTriggerL=0,
			bTriggerR=0,
		)
		vigem.DS4_REPORT_INIT(self._report)
		self._create_device()

	def send(self, key, value):
		self._report.wButtons = 0
		if key in self.buttons:
			if value:
				self._wButtons.add(self.buttons[key])
			else:
				self._wButtons.remove(self.buttons[key])
		elif key in self.specials:
			if value:
				self._report.bSpecial |= self.specials[key]
				logger.debug(f'{self.specials[key].name}::{self.specials[key]}')
			else:
				self._report.bSpecial = 0
				logger.debug(f'{self.specials[key].name}::0')
		elif key in self.dpad:
			if value:
				vigem.DS4_SET_DPAD(self._report, self.dpad[key])
				logger.debug(f'{self.dpad[key].name}')
			else:
				vigem.DS4_SET_DPAD(
					self._report, vigem.DS4_DPAD_DIRECTIONS.DS4_BUTTON_DPAD_NONE)
				logger.debug(f'{vigem.DS4_DPAD_DIRECTIONS.DS4_BUTTON_DPAD_NONE.name}')
		elif key in self.triggers:
			if value:
				setattr(self._report, self.triggers[key][0], vigem.XUSB_TRIGGER_MAX)
				logger.debug(f'{self.triggers[key][0]}::{vigem.XUSB_TRIGGER_MAX}')
				self._wButtons.add(self.triggers[key][1])
			else:
				setattr(self._report, self.triggers[key][0], 0)
				logger.debug(f'{self.triggers[key][0]}::0')
				self._wButtons.remove(self.triggers[key][1])
		elif key in self.axes_vertical:
			axis = round(value * 127) + 127
			setattr(self._report, self.axes_vertical[key], axis)
			logger.debug(f'{self.axes_vertical[key]}::{axis}')
		elif key in self.axes_horizontal:
			axis = round(value * 127) + 127
			setattr(self._report, self.axes_horizontal[key], axis)
			logger.debug(f'{self.axes_horizontal[key]}::{axis}')
		self._report.wButtons |= reduce(lambda a, b: a | b, self._wButtons, 0)
		logger.debug(
			f'wButtons::Mask {self._report.wButtons}::\
				{[b.name for b in self._wButtons]}')
		error = vigem.VIGEM_ERRORS(
			vigem.target_ds4_update(self._client, self._target, self._report)
		)
		if error != vigem.VIGEM_ERRORS.VIGEM_ERROR_NONE:
			logger.error(
				f'Virtual {self.type} device {self.device} at {self.address}::\
					Updating device state failed::{error.name}'
			)



import sys
from enum import IntFlag, IntEnum
from ctypes import Structure, WinDLL
from ctypes import c_ushort, c_short, c_byte, c_void_p, c_uint
from pathlib import PurePath
from pkg_resources import resource_filename


if sys.maxsize > 2**32:
	ARCH = 'x64'
else:
	ARCH = 'x86'


XUSB_TRIGGER_MAX = 255
XUSB_THUMB_MAX = 32767


class VIGEM_TARGET_TYPE(IntFlag):
	Xbox360Wired    = 0
	XboxOneWired    = 1
	DualShock4Wired = 2


class XUSB_BUTTON(IntFlag):
	XUSB_GAMEPAD_DPAD_UP        = 0x0001
	XUSB_GAMEPAD_DPAD_DOWN      = 0x0002
	XUSB_GAMEPAD_DPAD_LEFT      = 0x0004
	XUSB_GAMEPAD_DPAD_RIGHT     = 0x0008
	XUSB_GAMEPAD_START          = 0x0010
	XUSB_GAMEPAD_BACK           = 0x0020
	XUSB_GAMEPAD_LEFT_THUMB     = 0x0040
	XUSB_GAMEPAD_RIGHT_THUMB    = 0x0080
	XUSB_GAMEPAD_LEFT_SHOULDER  = 0x0100
	XUSB_GAMEPAD_RIGHT_SHOULDER = 0x0200
	XUSB_GAMEPAD_GUIDE          = 0x0400
	XUSB_GAMEPAD_A              = 0x1000
	XUSB_GAMEPAD_B              = 0x2000
	XUSB_GAMEPAD_X              = 0x4000
	XUSB_GAMEPAD_Y              = 0x8000


class DS4_BUTTONS(IntFlag):
	DS4_BUTTON_THUMB_RIGHT      = 1 << 15
	DS4_BUTTON_THUMB_LEFT       = 1 << 14
	DS4_BUTTON_OPTIONS          = 1 << 13
	DS4_BUTTON_SHARE            = 1 << 12
	DS4_BUTTON_TRIGGER_RIGHT    = 1 << 11
	DS4_BUTTON_TRIGGER_LEFT     = 1 << 10
	DS4_BUTTON_SHOULDER_RIGHT   = 1 << 9
	DS4_BUTTON_SHOULDER_LEFT    = 1 << 8
	DS4_BUTTON_TRIANGLE         = 1 << 7
	DS4_BUTTON_CIRCLE           = 1 << 6
	DS4_BUTTON_CROSS            = 1 << 5
	DS4_BUTTON_SQUARE           = 1 << 4


class DS4_SPECIAL_BUTTONS(IntFlag):
	DS4_SPECIAL_BUTTON_PS       = 1 << 0
	DS4_SPECIAL_BUTTON_TOUCHPAD = 1 << 1


class DS4_DPAD_DIRECTIONS(IntEnum):
	DS4_BUTTON_DPAD_NONE        = 0x8
	DS4_BUTTON_DPAD_NORTHWEST   = 0x7
	DS4_BUTTON_DPAD_WEST        = 0x6
	DS4_BUTTON_DPAD_SOUTHWEST   = 0x5
	DS4_BUTTON_DPAD_SOUTH       = 0x4
	DS4_BUTTON_DPAD_SOUTHEAST   = 0x3
	DS4_BUTTON_DPAD_EAST        = 0x2
	DS4_BUTTON_DPAD_NORTHEAST   = 0x1
	DS4_BUTTON_DPAD_NORTH       = 0x0


class XUSB_REPORT(Structure):
	_fields_ = (
		('wButtons', c_ushort),
		('bLeftTrigger', c_byte),
		('bRightTrigger', c_byte),
		('sThumbLX', c_short),
		('sThumbLY', c_short),
		('sThumbRX', c_short),
		('sThumbRY', c_short),
	)


class DS4_REPORT(Structure):
	_fields_ = (
		('bThumbLX', c_byte),
		('bThumbLY', c_byte),
		('bThumbRX', c_byte),
		('bThumbRY', c_byte),
		('wButtons', c_ushort),
		('bSpecial', c_byte),
		('bTriggerL', c_byte),
		('bTriggerR', c_byte),
	)


def DS4_SET_DPAD(report, direction):
	report.wButtons &= ~0xF
	report.wButtons |= direction


def DS4_REPORT_INIT(report):
	report.bThumbLX = 0x80
	report.bThumbLY = 0x80
	report.bThumbRX = 0x80
	report.bThumbRY = 0x80
	DS4_SET_DPAD(report, DS4_DPAD_DIRECTIONS.DS4_BUTTON_DPAD_NONE)


class VIGEM_ERRORS(IntEnum):
	VIGEM_ERROR_NONE                        = 0x20000000
	VIGEM_ERROR_BUS_NOT_FOUND               = 0xE0000001
	VIGEM_ERROR_NO_FREE_SLOT                = 0xE0000002
	VIGEM_ERROR_INVALID_TARGET              = 0xE0000003
	VIGEM_ERROR_REMOVAL_FAILED              = 0xE0000004
	VIGEM_ERROR_ALREADY_CONNECTED           = 0xE0000005
	VIGEM_ERROR_TARGET_UNINITIALIZED        = 0xE0000006
	VIGEM_ERROR_TARGET_NOT_PLUGGED_IN       = 0xE0000007
	VIGEM_ERROR_BUS_VERSION_MISMATCH        = 0xE0000008
	VIGEM_ERROR_BUS_ACCESS_FAILED           = 0xE0000009
	VIGEM_ERROR_CALLBACK_ALREADY_REGISTERED = 0xE0000010
	VIGEM_ERROR_CALLBACK_NOT_FOUND          = 0xE0000011
	VIGEM_ERROR_BUS_ALREADY_CONNECTED       = 0xE0000012
	VIGEM_ERROR_BUS_INVALID_HANDLE          = 0xE0000013
	VIGEM_ERROR_XUSB_USERINDEX_OUT_OF_RANGE = 0xE0000014


if getattr(sys, 'frozen', False) and hasattr(sys, '_MEIPASS'):
	CLIENT = WinDLL(str(PurePath(sys._MEIPASS, ARCH, 'ViGEmClient.dll')))
else:
	CLIENT = WinDLL(resource_filename(f'{__package__}.{ARCH}', 'ViGEmClient.dll'))


alloc = CLIENT.vigem_alloc
alloc.argtypes = ()
alloc.restype = c_void_p


free = CLIENT.vigem_free
free.argtypes = (c_void_p,)
free.restype = None


connect = CLIENT.vigem_connect
connect.argtypes = (c_void_p,)
connect.restype = c_uint


disconnect = CLIENT.vigem_disconnect
disconnect.argtypes = (c_void_p,)
disconnect.restype = None


target_x360_alloc = CLIENT.vigem_target_x360_alloc
target_x360_alloc.argtypes = ()
target_x360_alloc.restype = c_void_p


target_ds4_alloc = CLIENT.vigem_target_ds4_alloc
target_ds4_alloc.argtypes = ()
target_ds4_alloc.restype = c_void_p


target_free = CLIENT.vigem_target_free
target_free.argtypes = (c_void_p,)
target_free.restype = None


target_add = CLIENT.vigem_target_add
target_add.argtypes = (c_void_p, c_void_p)
target_add.restype = c_uint


target_remove = CLIENT.vigem_target_remove
target_remove.argtypes = (c_void_p, c_void_p)
target_remove.restype = c_uint


target_x360_update = CLIENT.vigem_target_x360_update
target_x360_update.argtypes = (c_void_p, c_void_p, XUSB_REPORT)
target_x360_update.restype = c_uint


target_ds4_update = CLIENT.vigem_target_ds4_update
target_ds4_update.argtypes = (c_void_p, c_void_p, DS4_REPORT)
target_ds4_update.restype = c_uint



import logging
from abc import ABC, abstractmethod
from evdev import UInput, AbsInfo, ecodes as e


logger = logging.getLogger('J2DX.Linux')


class Device(ABC):

	@abstractmethod
	def __init__(self, device, addr):
		self.device = device
		self.address = addr

	def close(self):
		self._ui.close()
		logger.debug(
			f'Destroyed virtual {self.type} device for {self.device} \
				at {self.address}')

	@abstractmethod
	def send(self, key, value):
		pass


class X360Device(Device):

	capabilities = {
		e.EV_KEY: [
			e.BTN_Y,
			e.BTN_X,
			e.BTN_A,
			e.BTN_B,
			e.BTN_THUMBL,
			e.BTN_THUMBR,
			e.BTN_SELECT,
			e.BTN_START,
			e.BTN_MODE,
			e.BTN_TRIGGER_HAPPY1,
			e.BTN_TRIGGER_HAPPY2,
			e.BTN_TRIGGER_HAPPY3,
			e.BTN_TRIGGER_HAPPY4,
			e.BTN_TL,
			e.BTN_TL2,
			e.BTN_TR,
			e.BTN_TR2,
		],
		e.EV_ABS: [
			(e.ABS_X, AbsInfo(value=0, min=0, max=255, fuzz=0, flat=0, resolution=0)),
			(e.ABS_Y, AbsInfo(value=0, min=0, max=255, fuzz=0, flat=0, resolution=0)),
			(e.ABS_RX, AbsInfo(value=0, min=0, max=255, fuzz=0, flat=0, resolution=0)),
			(e.ABS_RY, AbsInfo(value=0, min=0, max=255, fuzz=0, flat=0, resolution=0)),
		],
	}
	buttons = {
		'main-button': e.BTN_MODE,
		'back-button': e.BTN_SELECT,
		'start-button': e.BTN_START,
		'left-stick-press': e.BTN_THUMBL,
		'right-stick-press': e.BTN_THUMBR,
		'left-bumper': e.BTN_TL,
		'left-trigger': e.BTN_TL2,
		'right-bumper': e.BTN_TR,
		'right-trigger': e.BTN_TR2,
		'up-button': e.BTN_TRIGGER_HAPPY3,
		'right-button': e.BTN_TRIGGER_HAPPY2,
		'down-button': e.BTN_TRIGGER_HAPPY4,
		'left-button': e.BTN_TRIGGER_HAPPY1,
		'y-button': e.BTN_Y,
		'x-button': e.BTN_X,
		'a-button': e.BTN_A,
		'b-button': e.BTN_B,
	}
	axes = {
		'left-stick-X': e.ABS_X,
		'left-stick-Y': e.ABS_Y,
		'right-stick-X': e.ABS_RX,
		'right-stick-Y': e.ABS_RY,
	}

	def __init__(self, device, addr):
		super().__init__(device, addr)
		self.type = "Xbox 360 Controller"
		self._ui = UInput(
			events=self.capabilities,
			name=self.type,
			vendor=0x045e,
			product=0x028e,
			version=0x0110,
			bustype=e.BUS_USB,
		)

	def send(self, key, value):
		if key in self.buttons:
			logger.debug(f'Sending button event::{e.keys[self.buttons[key]]}: {value}')
			self._ui.write(e.EV_KEY, self.buttons[key], value)
			self._ui.syn()
		elif key in self.axes:
			coord = round(127 * value) + 127
			logger.debug(f'Sending axis event::{e.ABS[self.axes[key]]}: {coord}')
			self._ui.write(e.EV_ABS, self.axes[key], coord)
			self._ui.syn()


class DS4Device(Device):

	capabilities = {
		e.EV_KEY: [
			e.BTN_WEST,    # Square
			e.BTN_SOUTH,   # Cross
			e.BTN_EAST,    # Circle
			e.BTN_NORTH,   # Triangle
			e.BTN_TL,      # L1
			e.BTN_TR,      # R1
			e.BTN_TL2,     # L2
			e.BTN_TR2,     # R2
			e.BTN_SELECT,  # Share
			e.BTN_START,   # Options
			e.BTN_THUMBL,  # L3
			e.BTN_THUMBR,  # R3
			e.BTN_MODE,    # PS
		],
		e.EV_ABS: [
			(e.ABS_X, AbsInfo(
				value=0, min=0, max=255, fuzz=0, flat=5, resolution=0)),
			(e.ABS_Y, AbsInfo(
				value=0, min=0, max=255, fuzz=0, flat=5, resolution=0)),
			# right stick X
			(e.ABS_RX, AbsInfo(
				value=0, min=0, max=255, fuzz=0, flat=5, resolution=0)),
			# right stick Y
			(e.ABS_RY, AbsInfo(
				value=0, min=0, max=255, fuzz=0, flat=5, resolution=0)),
			# L2
			(e.ABS_Z, AbsInfo(
				value=0, min=0, max=255, fuzz=0, flat=5, resolution=0)),
			# R2
			(e.ABS_RZ, AbsInfo(
				value=0, min=0, max=255, fuzz=0, flat=5, resolution=0)),
			# dpad_left, dpad_right
			(e.ABS_HAT0X, AbsInfo(
				value=0, min=0, max=255, fuzz=0, flat=0, resolution=0)),
			# dpad_up, dpad_down
			(e.ABS_HAT0Y, AbsInfo(
				value=0, min=0, max=255, fuzz=0, flat=0, resolution=0)),
		]
	}
	buttons = {
		'main-button': e.BTN_MODE,
		'back-button': e.BTN_SELECT,
		'start-button': e.BTN_START,
		'left-stick-press': e.BTN_THUMBL,
		'right-stick-press': e.BTN_THUMBR,
		'left-bumper': e.BTN_TL,
		'left-trigger': e.BTN_TL2,
		'right-bumper': e.BTN_TR,
		'right-trigger': e.BTN_TR2,
		'y-button': e.BTN_NORTH,
		'x-button': e.BTN_EAST,
		'a-button': e.BTN_SOUTH,
		'b-button': e.BTN_WEST,
	}
	dpad = {
		'up-button': e.ABS_HAT0Y,
		'right-button': e.ABS_HAT0X,
		'down-button': e.ABS_HAT0Y,
		'left-button': e.ABS_HAT0X,
	}
	axes = {
		'left-stick-X': e.ABS_X,
		'left-stick-Y': e.ABS_Y,
		'right-stick-X': e.ABS_RX,
		'right-stick-Y': e.ABS_RY,
	}

	def __init__(self, device, addr):
		super().__init__(device, addr)
		self.type = "Sony Computer Entertainment Wireless Controller"
		self._ui = UInput(
			events=self.capabilities,
			name=self.type,
			vendor=1356,
			product=1476,
			version=273,
			bustype=e.BUS_USB,
		)

	def send(self, key, value):
		if key in self.buttons:
			logger.debug(f'Sending button event::{e.keys[self.buttons[key]]}: {value}')
			self._ui.write(e.EV_KEY, self.buttons[key], value)
			self._ui.syn()
		elif key in self.dpad:
			if key in {'up-button', 'left-button'}:
				value = 0 if value else 127
			else:
				value = 255 if value else 127
			logger.debug(f'Sending axis event::{e.ABS[self.dpad[key]]}: {value}')
			self._ui.write(e.EV_ABS, self.dpad[key], value)
			self._ui.syn()
		elif key in self.axes:
			coord = round(127 * value) + 127
			logger.debug(f'Sending axis event::{e.ABS[self.axes[key]]}: {coord}')
			self._ui.write(e.EV_ABS, self.axes[key], coord)
			self._ui.syn()



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['j2dx', 'j2dx.win.ViGEm.x64', 'j2dx.win.ViGEm.x86']

package_data = \
{'': ['*'], 'j2dx': ['nix/*', 'win/*', 'win/ViGEm/*']}

install_requires = \
['eventlet>=0.25,<0.26', 'python-socketio>=4.2,<5.0', 'qrcode>=6.1,<7.0']

extras_require = \
{':sys_platform == "linux"': ['evdev>=1.2,<2.0'],
 ':sys_platform == "win32"': ['colorama>=0.4,<0.5']}

entry_points = \
{'console_scripts': ['j2dx = j2dx:main']}

setup_kwargs = {
    'name': 'j2dx',
    'version': '0.1.0',
    'description': 'Use your Android phone as a virtual Xbox 360 controller or DualShock 4 gamepad on your Windows or Linux PC. This is the server that runs on Windows or Linux.',
    'long_description': "# Joy2DroidX\n\n![Joy2DroidX](assets/screenshot.png)\n\nJoy2DroidX allows you to use your Android device as a\nvirtual Xbox 360 controller or DualShock 4 gamepad.\n\nIt consists of a server that runs on Windows and Linux\nand an Android app. While there's nothing preventing the\napp from running on any Android device, it's been only\ntested on Android phones.\n\n\n### Server\n\nThe server (this app) listens for input from connected\nAndroid devices and manages creation/deletion of virtual devices.\nIt uses UInput on Linux and [ViGEm](https://github.com/ViGEm) on Windows.\n\nWhile running the server *does not* require any special\nprivileges, the initial setup (setting UInput permissions on Linux and installing driver on Windows) *requires root/administrator* access.\n\n### Client\n\nYou can find more information about the Android app as well the sources [here](https://github.com/OzymandiasTheGreat/Joy2DroidX).\n\n\n## Installation\n\nJoy2DroidX is distributed as a portable app on Windows and as an AppImage on Linux.\nYou can find latest versions on [releases page](https://github.com/OzymandiasTheGreat/Joy2DroidX-server/releases).\n\nAlternatively, if you have python 3 and pip setup, you can install from [pypi](https://pypi.org/project/j2dx/):\n\n```\npip install j2dx\n```\n\n\n## Usage\n\n### First run\n\nYou need to setup the system before the first run.\nJoy2DroidX provides a convenience command that does this for you, it however requires root/administrator access.\n\nJust run `j2dx --setup` as root or from administrator command prompt.\n\nOn Linux this will create a udev rule for UInput and add current user to `j2dx` group. If you're not using sudo or user detection fails for another reason, you can provide username as an argument to `--setup`.\nFor udev rules and group changes to take effect you'll have to restart your system.\n\nOn Windows this will download [ViGEmBus driver](https://github.com/ViGEm/ViGEmBus) and prompt you to install it.\nOnce the driver is setup you can use Joy2DroidX, no restart necessary.\n\n### Regular usage\n\nRun `j2dx` (on windows you can just double click `j2dx.exe`), scan QRCode from the Android app and that's it.\nEverything should just work. Switching device mode is done from the Android app.\n\nThe server should not need any extra configuration.\nIf you have an unsual network setup or default port is used by another process, there are a couple option you can modify:\n\n- `-p, --port` allows you to use a different port. Default is 8013.\n- `-H, --host` if hostname detection fails you can specify a hostname or your computers IP address.\n- `-d, --debug` you shouldn't need this one. If you do encounter bugs, run `j2dx -d` and open an issue with a link to debug output (use a gist or pastebin for this).\n",
    'author': 'Tomas Ravinskas',
    'author_email': 'tomas.rav@gmail.com',
    'url': 'https://github.com/OzymandiasTheGreat/Joy2DroidX-server',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
